/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)

  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
                            funkyass (funkyass@spam.shaw.ca),
                            Joel Yliluoma (http://iki.fi/bisqwit/)
                            Kris Bleakley (codeviolation@hotmail.com),
                            Matthew Kendora,
                            Nach (n-a-c-h@users.sourceforge.net),
                            Peter Bortas (peter@bortas.org) and
                            zones (kasumitokoduck@yahoo.com)

  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and Nach

  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman, neviksti (neviksti@hotmail.com),
                            Kris Bleakley, Andreas Naive

  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
                            Kris Bleakley
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-DD1 C emulator code
  (c) Copyright 2003 Brad Jorsch with research by
                     Andreas Naive and John Weidman
 
  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  ST010 C++ emulator code
  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora

  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman


  SH assembler code partly based on x86 assembler code
  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 

 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

/*******************************************************************************
  Snes9x for Mac OS X by zones
  
  This build was started from John Stiles' port.
*******************************************************************************/

#include "port.h"

#include "mac-prefix.h"
#include "mac-os.h"
#include "mac-snes9x.h"
#include "mac-appleevent.h"

static pascal OSErr AEoapp(const AppleEvent *, AppleEvent *, long);
static pascal OSErr AErapp(const AppleEvent *, AppleEvent *, long);
static pascal OSErr AEpdoc(const AppleEvent *, AppleEvent *, long);
static pascal OSErr AEquit(const AppleEvent *, AppleEvent *, long);
static pascal OSErr AEodoc(const AppleEvent *, AppleEvent *, long);

static AEEventHandlerUPP oappUPP, rappUPP, pdocUPP, quitUPP, odocUPP;

static pascal OSErr AEoapp(const AppleEvent *theEvent, AppleEvent *theReply, long refCon)
{
	#pragma unused (theEvent, theReply, refCon)
	
	if (running)
		return noErr;
	
	if (startopendlog)
	{
		if (SNES9X_OpenCart(nil))
		{
			SNES9X_Go();
			QuitApplicationEventLoop();
		}
	}
	
	return noErr;
}

static pascal OSErr AErapp(const AppleEvent *theEvent, AppleEvent *theReply, long refCon)
{
	#pragma unused (theEvent, theReply, refCon)
	
	if (running)
		return noErr;

	if (!cartOpen)
	{
		if (startopendlog)
		{
			if (SNES9X_OpenCart(nil))
			{
				SNES9X_Go();
				QuitApplicationEventLoop();
			}
		}
	}
	else
	{
		SNES9X_Go();
		QuitApplicationEventLoop();
	}
	
	return noErr;
}

static pascal OSErr AEpdoc(const AppleEvent *theEvent, AppleEvent *theReply, long refCon)
{
	#pragma unused (theEvent, theReply, refCon)
	
	return errAEEventNotHandled;
}

static pascal OSErr AEquit(const AppleEvent *theEvent, AppleEvent *theReply, long refCon)
{	
	#pragma unused (theEvent, theReply, refCon)
	
	if (running)
		return noErr;

	SNES9X_Quit();
	QuitApplicationEventLoop();
	
	return noErr;
}

static pascal OSErr AEodoc(const AppleEvent *theEvent, AppleEvent *theReply, long refCon)
{
	OSErr 		err;
	FSRef		ref;
	AEDescList	docList;
	AEKeyword	keywd;
	DescType	rtype;
	Size		acsize;
	long		count;
	
	#pragma unused (theReply, refCon)
	
	if (running)
		return noErr;

	err = AEGetParamDesc(theEvent, keyDirectObject, typeAEList, &docList);
	if (err)
		return noErr;
	
	err = AECountItems(&docList, &count);
	if (err || (count != 1))
	{
		err = AEDisposeDesc(&docList);
		return noErr;
	}
	
	err = AEGetNthPtr(&docList, 1, typeFSRef, &keywd, &rtype, &ref, sizeof(FSRef), &acsize);
	if (err == noErr)
	{
		if (SNES9X_OpenCart(&ref))
		{
			SNES9X_Go();
			QuitApplicationEventLoop();
		}
	}

	err = AEDisposeDesc(&docList);

	return noErr;
}

void InitAppleEvents(void)
{
	OSErr	err;
	
	oappUPP = NewAEEventHandlerUPP(AEoapp);
	err = AEInstallEventHandler(kCoreEventClass, kAEOpenApplication,   oappUPP, 0L, false);
	
	odocUPP = NewAEEventHandlerUPP(AEodoc);
	err = AEInstallEventHandler(kCoreEventClass, kAEOpenDocuments,     odocUPP, 0L, false);
		
	pdocUPP = NewAEEventHandlerUPP(AEpdoc);
	err = AEInstallEventHandler(kCoreEventClass, kAEPrintDocuments,    pdocUPP, 0L, false);
		
	quitUPP = NewAEEventHandlerUPP(AEquit);
	err = AEInstallEventHandler(kCoreEventClass, kAEQuitApplication,   quitUPP, 0L, false);
		
	rappUPP = NewAEEventHandlerUPP(AErapp);
	err = AEInstallEventHandler(kCoreEventClass, kAEReopenApplication, rappUPP, 0L, false);
}

void DeinitAppleEvents(void)
{
	DisposeAEEventHandlerUPP(oappUPP);
	DisposeAEEventHandlerUPP(odocUPP);
	DisposeAEEventHandlerUPP(pdocUPP);
	DisposeAEEventHandlerUPP(quitUPP);
	DisposeAEEventHandlerUPP(rappUPP);
}
