/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)

  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
                            funkyass (funkyass@spam.shaw.ca),
                            Joel Yliluoma (http://iki.fi/bisqwit/)
                            Kris Bleakley (codeviolation@hotmail.com),
                            Matthew Kendora,
                            Nach (n-a-c-h@users.sourceforge.net),
                            Peter Bortas (peter@bortas.org) and
                            zones (kasumitokoduck@yahoo.com)

  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and Nach

  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman, neviksti (neviksti@hotmail.com),
                            Kris Bleakley, Andreas Naive

  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
                            Kris Bleakley
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-DD1 C emulator code
  (c) Copyright 2003 Brad Jorsch with research by
                     Andreas Naive and John Weidman
 
  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  ST010 C++ emulator code
  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora

  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman


  SH assembler code partly based on x86 assembler code
  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 

 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

/*******************************************************************************
  Snes9x for Mac OS X by zones
  
  This build was started from John Stiles' port.
*******************************************************************************/

#include "memmap.h"
#include "apu.h"
#include "snapshot.h"
#include "soundux.h"

#include "mac-prefix.h"
#include "mac-audio.h"
#include "mac-dialog.h"
#include "mac-file.h"
#include "mac-os.h"
#include "mac-musicbox.h"
#include "mac-stringtools.h"

#define	mbxBarWidth		12.0
#define	mbxBarHeight	128.0
#define	mbxBarSpace		2.0
#define	mbxLRSpace		20.0
#define	mbxRightBarX	(mbxLRSpace + (mbxBarWidth * 8.0 + mbxBarSpace * 7.0)) 
#define yyscale			(128.0 / (sqrtf(64.0)))

extern char gMacRomName[ROM_NAME_LEN];

volatile Boolean	mboxPause = false;
uint16				mscbx = ~0;

static void SPCPlayExec(void);
static void SPCPlayFreeze(void);
static void SPCPlayDefrost(void);
static void MusicBoxForceFreeze(void);
static void MusicBoxForceDefrost(void);
static void MusicBoxInitIndicator(void);
static pascal void SoundTask(TMTask *);
static pascal void MusicBoxDrawIndicator(EventLoopTimerRef, void *);
static pascal OSStatus MusicBoxEventHandler(EventHandlerCallRef, EventRef, void *);

static volatile Boolean	installed, stopNow, showIndicator, mbxFinished;

static int32			mbxVCounter;
static short			mbxOpenedHeight, mbxClosedHeight;
static float			mbxOffsetX, mbxOffsetY, mbxMarginX, mbxMarginY;
static float			mbxViewWidth, mbxViewHeight;
static long				delay;
static CGContextRef		mboxctx;

static SAPU				*StoredAPU;
static SAPURegisters	*StoredAPURegisters;
static SSoundData		*StoredSoundData;
static uint8			*StoredIAPURAM;

static short			prevLMax[8], prevRMax[8];
static short			prevLVol[8], prevRVol[8];
static long long		barTimeL[8], barTimeR[8];

void MusicBoxDialog(void)
{
	OSStatus	err;
	IBNibRef	nibRef;

	if (!cartOpen)
		return;
	
	err = CreateNibReference(kMacS9XCFString, &nibRef);
	if (err == noErr)
	{
		CFURLRef	iconURL;
		FSRef		iconFSRef;
		IconRef		actIcon;
		WindowRef	tWindowRef;
		
		actIcon = nil;

		if (musicboxmode == kMBXSoundEmulation)
			iconURL = CFBundleCopyResourceURL(CFBundleGetMainBundle(), CFSTR("ledoff"), CFSTR("icns"), nil);
		else
			iconURL = CFBundleCopyResourceURL(CFBundleGetMainBundle(), CFSTR("ledon" ), CFSTR("icns"), nil);

		if (iconURL)
		{
			if (CFURLGetFSRef(iconURL, &iconFSRef))
				err = RegisterIconRefFromFSRef('~9X~', 'micn', &iconFSRef, &actIcon);
					
			CFRelease(iconURL);
		}

		err = CreateWindowFromNib(nibRef, CFSTR("MusicBox"), &tWindowRef);
		if (err == noErr)
		{
			EventHandlerRef		eref;
			EventHandlerUPP		eventUPP;
			EventLoopTimerRef	mboxTimerRef;
			EventLoopTimerUPP	mboxTimerUPP;
			EventTypeSpec		windowEvents[] = { { kEventClassCommand, kEventCommandProcess } };
			TMTask				tmtask;
			Rect				rct, tmp_rct, ind_rct;
			CFStringRef			sref;
			ControlRef			control;
			ControlID			controlID;

			mboxPause = false;
			mbxFinished = false;

			switch (drawingMethod)
			{
				case kDrawingOpenGL:
					Settings.OpenGLEnable = true;
					break;
					
				case kDrawingDirect:
				case kDrawingBlitGL:
					Settings.OpenGLEnable = false;
			}
			
			// 107's enhanced SPC player
			
			controlID.id = 0;

			if (musicboxmode == kMBXSoundEmulation)
			{
				controlID.signature = 'HEAD';
				GetControlByID(tWindowRef, &controlID, &control);
				EnableControl(control);

				StoredAPU          = new SAPU;
				StoredAPURegisters = new SAPURegisters;
				StoredSoundData    = new SSoundData;
				StoredIAPURAM      = new uint8 [0x10000];
				
				SPCPlayFreeze();
			}
			else
				MusicBoxForceFreeze();

			controlID.signature = 'LCD_';
			GetControlByID(tWindowRef, &controlID, &control);
			sref = CopyFixNameStrings(gMacRomName, Memory.ROMRegion);
			if (sref)
			{
				SetControlTitleWithCFString(control, sref);
				CFRelease(sref);
			}

			if (!useV2)
			{
				controlID.signature = 'S_EF';
				GetControlByID(tWindowRef, &controlID, &control);
				DeactivateControl(control);
			}

			for (int i = 0; i < 5; i++)
				controlPad[i] = 0;

			showIndicator = false;
	
			eventUPP = NewEventHandlerUPP(MusicBoxEventHandler);
			err = InstallWindowEventHandler(tWindowRef, eventUPP, 1, windowEvents, (void *) tWindowRef, &eref);
			
			mboxTimerUPP = NewEventLoopTimerUPP(MusicBoxDrawIndicator);
			err = InstallEventLoopTimer(GetCurrentEventLoop(), kEventDurationNoWait, kEventDurationSecond / (double) Memory.ROMFramesPerSecond, mboxTimerUPP, (void *) tWindowRef, &mboxTimerRef);

			tmtask.tmWakeUp   = 0;
			tmtask.tmReserved = 0;
			tmtask.tmAddr     = NewTimerUPP(SoundTask);			
			err = InstallXTimeTask((QElemPtr) &tmtask);

			delay = 1000000 / Memory.ROMFramesPerSecond;
			stopNow = false;
			installed = true;

			mbxVCounter = 0;
			
			mscbx = ~0;
			MusicBoxInitIndicator();

			CreateCGContextForPort(GetWindowPort(tWindowRef), &mboxctx);

			controlID.signature = 'indc';			
			GetControlByID(tWindowRef, &controlID, &control);
			GetControlBounds(control, &ind_rct);

			controlID.signature = 'tr_i';			
			GetControlByID(tWindowRef, &controlID, &control);
			GetControlBounds(control, &tmp_rct);
			mbxClosedHeight = tmp_rct.bottom + 11;			
			
			GetWindowBounds(tWindowRef, kWindowContentRgn, &rct);
			mbxOpenedHeight = rct.bottom - rct.top;
			mbxOffsetX = ind_rct.left;
			mbxOffsetY = rct.bottom - rct.top - ind_rct.bottom;
			mbxMarginY = (float) (ind_rct.bottom - ind_rct.top  - mbxBarHeight) / 2.0;
			mbxMarginX = (float) (ind_rct.right  - ind_rct.left - ((mbxBarWidth * 8.0 + mbxBarSpace * 7.0) * 2.0 + mbxLRSpace)) / 2.0;
			
			mbxViewWidth  = ind_rct.right  - ind_rct.left;
			mbxViewHeight = ind_rct.bottom - ind_rct.top;

			rct.bottom = rct.top + mbxClosedHeight;
			SetWindowBounds(tWindowRef, kWindowContentRgn, &rct);
			SyncCGContextOriginWithPort(mboxctx, GetWindowPort(tWindowRef));
			
			MacStartSound();

			err = PrimeTimeTask((QElemPtr) &tmtask, -delay);
			
			MoveWindowPosition(tWindowRef, kWindowMusicBox);
			ShowWindow(tWindowRef);						
			err = RunAppModalLoopForWindow(tWindowRef);						
			HideWindow(tWindowRef);
			SaveWindowPosition(tWindowRef, kWindowMusicBox);
			
			CGContextRelease(mboxctx);
			
			stopNow = true;
			while (installed)
				MPYield();
			
			MacStopSound();
			
			err = RemoveTimeTask((QElemPtr) &tmtask);
			DisposeTimerUPP(tmtask.tmAddr);
			
			err = RemoveEventLoopTimer(mboxTimerRef);
			DisposeEventLoopTimerUPP(mboxTimerUPP);
			
			err = RemoveEventHandler(eref);
			DisposeEventHandlerUPP(eventUPP);
			
			ReleaseWindow(tWindowRef);
			
			mscbx = ~0;
			
			mbxFinished = true;
			
			if (musicboxmode == kMBXSoundEmulation)
			{
 				SPCPlayDefrost();

				delete    StoredAPU;
				delete    StoredAPURegisters;
				delete    StoredSoundData;
				delete [] StoredIAPURAM;				
			}
			else
				MusicBoxForceDefrost();

			Settings.OpenGLEnable = false;
		}

		if (actIcon)
			err = UnregisterIconRef('~9X~', 'micn');
		
		DisposeNibReference(nibRef);
	}
}

static pascal OSStatus MusicBoxEventHandler(EventHandlerCallRef inHandlerCallRef, EventRef inEvent, void *inUserData)
{
    #pragma unused (inHandlerCallRef)
    
    OSStatus	err, result = eventNotHandledErr;
	WindowRef	tWindowRef = (WindowRef) inUserData;

	switch (GetEventClass(inEvent))
	{
		case kEventClassCommand:
			switch (GetEventKind(inEvent))
			{
				case kEventCommandProcess:
					HICommand	cmd;

					err = GetEventParameter(inEvent, kEventParamDirectObject, typeHICommand, nil, sizeof(HICommand), nil, &cmd);
					if (err == noErr)
					{
						Rect	rct;

						switch (cmd.commandID)
						{
							case 'bar1': mscbx ^= (1 <<  0);	result = noErr;	break;
							case 'bar2': mscbx ^= (1 <<  1);	result = noErr;	break;
							case 'bar3': mscbx ^= (1 <<  2);	result = noErr;	break;
							case 'bar4': mscbx ^= (1 <<  3);	result = noErr;	break;
							case 'bar5': mscbx ^= (1 <<  4);	result = noErr;	break;
							case 'bar6': mscbx ^= (1 <<  5);	result = noErr;	break;
							case 'bar7': mscbx ^= (1 <<  6);	result = noErr;	break;
							case 'bar8': mscbx ^= (1 <<  7);	result = noErr;	break;
							case 'bar9': mscbx ^= (1 <<  8);	result = noErr;	break;
							case 'bara': mscbx ^= (1 <<  9);	result = noErr;	break;
							case 'barb': mscbx ^= (1 << 10);	result = noErr;	break;
							case 'barc': mscbx ^= (1 << 11);	result = noErr;	break;
							case 'bard': mscbx ^= (1 << 12);	result = noErr;	break;
							case 'bare': mscbx ^= (1 << 13);	result = noErr;	break;
							case 'barf': mscbx ^= (1 << 14);	result = noErr;	break;
							case 'bar0': mscbx ^= (1 << 15);	result = noErr;	break;

							case 'Paus': 
								mboxPause = !mboxPause;
								S9xSetSoundMute(mboxPause);
								result = noErr;
								break;

							case 'tr_i':
								showIndicator = !showIndicator;
								GetWindowBounds(tWindowRef, kWindowContentRgn, &rct);								 
								rct.bottom = rct.top + (showIndicator ? mbxOpenedHeight : mbxClosedHeight);
								SetWindowBounds(tWindowRef, kWindowContentRgn, &rct);
								SyncCGContextOriginWithPort(mboxctx, GetWindowPort(tWindowRef));
								result = noErr;
								break;
						
							case 'ok  ':
								QuitAppModalLoopForWindow(tWindowRef);
								result = noErr;
								break;

							case 'HEAD':
								showIndicator = !showIndicator;
								SPCPlayDefrost();
								showIndicator = !showIndicator;
								result = noErr;
								break;
								
							case 'S_EF':
								HideWindow(tWindowRef);
								showIndicator = !showIndicator;
								ConfigureSoundEffects();
								showIndicator = !showIndicator;
								ShowWindow(tWindowRef);
								result = noErr;
						}
					}
			}
	}
	
	return result;
}

static pascal void SoundTask(TMTask *task)
{
	if (!stopNow)
	{
		if (!mboxPause)
			(musicboxmode == kMBXSoundEmulation) ? SPCPlayExec() : S9xMainLoop();

		PrimeTimeTask((QElemPtr) task, -delay);
	}
	else
		installed = false;
}

static void SPCPlayExec(void)
{
	for (;;)
	{	
		while (APU.Cycles < Settings.HBlankStart)
			APU_EXECUTE1();

		APU.Cycles -= Settings.H_Max;
	
		if (++mbxVCounter >= (Settings.PAL ? SNES_MAX_PAL_VCOUNTER : SNES_MAX_NTSC_VCOUNTER))
		{
			mbxVCounter = 0;
			break;
		}
		
		if (APU.TimerEnabled[2])		// 2: 64kHz
		{
			APU.Timer[2] += 4;
			while (APU.Timer[2] >= APU.TimerTarget[2])
			{
				IAPU.RAM[0xff] = (IAPU.RAM[0xff] + 1) & 0xf;
				APU.Timer[2] -= APU.TimerTarget[2];
				
				#ifdef SPC700_SHUTDOWN		
				IAPU.WaitCounter++;
				IAPU.APUExecuting = true;
				#endif		
			}
		}

		if (mbxVCounter & 1)
		{
			if (APU.TimerEnabled[0])	// 0: 8kHz
			{
				APU.Timer[0]++;
				if (APU.Timer[0] >= APU.TimerTarget[0])
				{
					IAPU.RAM[0xfd] = (IAPU.RAM[0xfd] + 1) & 0xf;
					APU.Timer[0] = 0;
					
					#ifdef SPC700_SHUTDOWN		
					IAPU.WaitCounter++;
					IAPU.APUExecuting = true;
					#endif		    
				}
			}
			
			if (APU.TimerEnabled[1])	// 1: 8kHz
			{
				APU.Timer[1]++;
				if (APU.Timer[1] >= APU.TimerTarget[1])
				{
					IAPU.RAM[0xfe] = (IAPU.RAM[0xfe] + 1) & 0xf;
					APU.Timer[1] = 0;

					#ifdef SPC700_SHUTDOWN		
					IAPU.WaitCounter++;
					IAPU.APUExecuting = true;
					#endif		    
				}
			}
		}
	}
	
	APURegisters.PC = IAPU.PC - IAPU.RAM;
    S9xAPUPackStatus();
}

static void MusicBoxForceFreeze(void)
{
	char	filename[PATH_MAX + 1];
	
	strcpy(filename, S9xGetFreezeFilename(999));
	strcat(filename, ".tmp");
	
	S9xFreezeGame(filename);
}

static void MusicBoxForceDefrost(void)
{
	char	filename[PATH_MAX + 1];
	
	strcpy(filename, S9xGetFreezeFilename(999));
	strcat(filename, ".tmp");
	
	S9xUnfreezeGame(filename);
	remove(filename);
}

static void SPCPlayFreeze(void)
{
	int	i, j;
		
	S9xSetSoundMute(true);
		
	for (i = 0; i < 8; i++)
    {
		SoundData.channels[i].previous16[0] = (int16) SoundData.channels[i].previous[0];
		SoundData.channels[i].previous16[1] = (int16) SoundData.channels[i].previous[1];
    }

	memcpy(StoredAPU,          &APU,          sizeof(SAPU));
	memcpy(StoredAPURegisters, &APURegisters, sizeof(SAPURegisters));
	memcpy(StoredSoundData,    &SoundData,    sizeof(SSoundData));
	memcpy(StoredIAPURAM,      IAPU.RAM,      0x10000);
	
	// Clear needless values : This may be meanless.
	
	StoredSoundData->no_filter = false;
	StoredSoundData->master_volume[0] = StoredSoundData->master_volume[1] = 0;
	StoredSoundData->echo_volume[0] = StoredSoundData->echo_volume[1] = 0;
	StoredSoundData->noise_hertz = 0;
	
	for (i = 0; i < 8; i++)
	{
		StoredSoundData->channels[i].frequency = 0;
		StoredSoundData->channels[i].block = nil;
		StoredSoundData->channels[i].echo_buf_ptr = nil;
		StoredSoundData->channels[i].envxx = 0;
		StoredSoundData->channels[i].next_sample = 0;
		StoredSoundData->channels[i].interpolate = 0;
		StoredSoundData->channels[i].previous[0] = StoredSoundData->channels[i].previous[1] = 0;
		//StoredSoundData->channels[i].last_valid_header = 0;		
		for (j = 0; j < 8; j++)
			StoredSoundData->channels[i].dummy[j] = 0;
	}
	
	S9xSetSoundMute(false);
}

static void SPCPlayDefrost(void)
{
    short   i;
	
	S9xResetAPU();
    S9xSetSoundMute(true);
	
	memcpy(&APU,          StoredAPU,          sizeof(SAPU));
	memcpy(&APURegisters, StoredAPURegisters, sizeof(SAPURegisters));
	memcpy(&SoundData,    StoredSoundData,    sizeof(SSoundData));
	memcpy(IAPU.RAM,      StoredIAPURAM,      0x10000);
	
	if (!mbxFinished)
	{
		for (i = 0; i < 4; i++)
			APU.OutPorts[i] = IAPU.RAM[0xf4 + i];
		
		//IAPU.TimerErrorCounter = 0;   // unused

		for (i = 0; i < 3; i++)
		{
			if (IAPU.RAM[0xfa + i] == 0)
				APU.TimerTarget[i] = 0x100;
			else
				APU.TimerTarget[i] = IAPU.RAM[0xfa + i];
		}

		S9xSetAPUControl(IAPU.RAM[0xf1] & 0xcf);   // Don't reset I/O ports
	}

	S9xSetSoundMute(false);

	IAPU.PC = IAPU.RAM + APURegisters.PC;
	S9xAPUUnpackStatus();
	
	if (APUCheckDirectPage())
	    IAPU.DirectPage = IAPU.RAM + 0x100;
	else
	    IAPU.DirectPage = IAPU.RAM;

	IAPU.APUExecuting = true;

	S9xFixSoundAfterSnapshotLoad();
}

static void MusicBoxInitIndicator(void)
{
	long long	currentTime;
	
	Microseconds((UnsignedWide *) &currentTime);
	
	for (int i = 0; i < 8; i++)
	{
		prevLMax[i] = prevRMax[i] = 0;
		prevLVol[i] = prevRVol[i] = 0;
		barTimeL[i] = barTimeR[i] = currentTime;
	}
}

static pascal void MusicBoxDrawIndicator(EventLoopTimerRef eref, void *userData)
{	
	#pragma unused (eref, userData)
	
	if (!showIndicator)
		return;
		
	static float	length[] = { 1.0, 1.0 };

	// Background
	
	CGRect	crt;
	
	CGContextSetLineWidth(mboxctx, 12.0);
	CGContextSetLineDash(mboxctx, 0, nil, 0);
	CGContextSetLineJoin(mboxctx, kCGLineJoinRound);

	crt = CGRectMake(mbxOffsetX + 8.0, mbxOffsetY + 6.0, mbxViewWidth - 16.0, mbxViewHeight - 16.0);
	CGContextSetRGBFillColor  (mboxctx, (214.0 / 256.0), (219.0 / 256.0), (191.0 / 256.0), 1.0);
	CGContextSetRGBStrokeColor(mboxctx, (214.0 / 256.0), (219.0 / 256.0), (191.0 / 256.0), 1.0);
	CGContextBeginPath(mboxctx);
	CGContextAddRect(mboxctx, crt);	
	CGContextDrawPath(mboxctx, kCGPathFillStroke);

	// Bar
	
	CGContextSetLineWidth(mboxctx, mbxBarWidth);
	CGContextSetLineDash(mboxctx, 0, length, 2);
	CGContextSetLineJoin(mboxctx, kCGLineJoinMiter);

	CGContextBeginPath(mboxctx);
	
	float   x = mbxOffsetX + mbxMarginX + mbxBarWidth / 2.0;
	for (int h = 0; h < 8; h++)
	{
		// Inactive
		
		CGContextSetRGBStrokeColor(mboxctx, (196.0 / 256.0), (200.0 / 256.0), (176.0 / 256.0), 1.0);
		
		CGContextMoveToPoint   (mboxctx, x,                mbxOffsetY + mbxMarginY);
		CGContextAddLineToPoint(mboxctx, x,                mbxOffsetY + mbxMarginY + mbxBarHeight);
		
		CGContextMoveToPoint   (mboxctx, x + mbxRightBarX, mbxOffsetY + mbxMarginY);
		CGContextAddLineToPoint(mboxctx, x + mbxRightBarX, mbxOffsetY + mbxMarginY + mbxBarHeight);
	
		CGContextStrokePath(mboxctx);
		
		// Max

		short		vl = SoundData.channels[h].left_vol_level;
		short		vr = SoundData.channels[h].right_vol_level;
		long long	currentTime;
		
		if (vl <= 0) vl = 0; else if (vl > 64) vl = 64; else vl = (short) (yyscale * sqrtf(vl)) & (~0 << 1);
		if (vr <= 0) vr = 0; else if (vr > 64) vr = 64; else vr = (short) (yyscale * sqrtf(vr)) & (~0 << 1);
	
		if (vl < prevLVol[h]) vl = ((prevLVol[h] + vl) >> 1);
		if (vr < prevRVol[h]) vr = ((prevRVol[h] + vr) >> 1);

		Microseconds((UnsignedWide *) &currentTime);
		
		// left

		if ((vl >= prevLMax[h]) && (vl > prevLVol[h]))
		{
			barTimeL[h] = currentTime;
			prevLMax[h] = vl;
		}
		else
		if ((prevLMax[h] > 0) && (barTimeL[h] + 1000000 > currentTime))
		{
			CGContextSetRGBStrokeColor(mboxctx, (22.0 / 256.0), (156.0 / 256.0), (20.0 / 256.0), (float) (barTimeL[h] + 1000000 - currentTime) / 1000000.0);
			
			CGContextMoveToPoint   (mboxctx, x, mbxOffsetY + mbxMarginY + (float) (prevLMax[h] - 2));
			CGContextAddLineToPoint(mboxctx, x, mbxOffsetY + mbxMarginY + (float) (prevLMax[h]    ));
			
			CGContextStrokePath(mboxctx);
		}
		else
			prevLMax[h] = 0;
		
		prevLVol[h] = vl;
		
		// right
		
		if ((vr >= prevRMax[h]) && (vr > prevRVol[h]))
		{
			barTimeR[h] = currentTime;
			prevRMax[h] = vr;
		}
		else
		if ((prevRMax[h] > 0) && (barTimeR[h] + 1000000 > currentTime))
		{
			CGContextSetRGBStrokeColor(mboxctx, (22.0 / 256.0), (156.0 / 256.0), (20.0 / 256.0), (float) (barTimeR[h] + 1000000 - currentTime) / 1000000.0);
			
			CGContextMoveToPoint   (mboxctx, x + mbxRightBarX, mbxOffsetY + mbxMarginY + (float) (prevRMax[h] - 2));
			CGContextAddLineToPoint(mboxctx, x + mbxRightBarX, mbxOffsetY + mbxMarginY + (float) (prevRMax[h]    ));
			
			CGContextStrokePath(mboxctx);
		}
		else
			prevRMax[h] = 0;
		
		prevRVol[h] = vr;

		// Active
		
		CGContextSetRGBStrokeColor(mboxctx, (22.0 / 256.0), (22.0 / 256.0), (20.0 / 256.0), 1.0);
		
		CGContextMoveToPoint   (mboxctx, x,                mbxOffsetY + mbxMarginY);
		CGContextAddLineToPoint(mboxctx, x,                mbxOffsetY + mbxMarginY + (float) vl);
		CGContextStrokePath(mboxctx);
		CGContextMoveToPoint   (mboxctx, x + mbxRightBarX, mbxOffsetY + mbxMarginY);
		CGContextAddLineToPoint(mboxctx, x + mbxRightBarX, mbxOffsetY + mbxMarginY + (float) vr);
		CGContextStrokePath(mboxctx);
		
		x += (mbxBarWidth + mbxBarSpace);
	}
		
	CGContextFlush(mboxctx);
}
